-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local stringLib = { }
local util 			= { }

npcg.functua.string = stringLib
npcg.functua.util 	= util

-------------------------------------------------------------------------------------
-- | String |------------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- searchF :: (string --> string) --> string --> [string] --> bool
-- Searches for the existence of a string after mapping a function over a list of strings.
function stringLib.searchF(f, s, xs)
  return functua.list.filter(
    function(x) return string.match(f(x), s) end,
    xs)
end

-- combine :: string -> string
-- Concatenates two strings even if either is nil.
function stringLib.combine(x,y)
  return (x or "") .. (y or "")
end

-- combineWithSpace :: string -> string
-- Concatenates two strings with a whitespace in between. If either is nil or empty, that one will be ignores.
function stringLib.combineWithSpace(x,y)
  if      x == nil or x == "" then  return y
  elseif  y == nil or y == "" then  return x
  else                              return x .. " " .. y end
end

-- nilOrWhitespace :: string -> bool
-- Returns whether a string is nil or empty.
function stringLib.nilOrWhitespace(s)
  return s == nil or s == ""
end

-- nilOrWhitespace :: string -> string -> bool
-- Returns whether a string is nil, empty or equal to the other string.
function stringLib.nilOrWhitespaceOr(s, t)
  return stringLib.nilOrWhitespace(s) or (s == t)
end

-- ucFirst :: string -> string
-- Capitalizes the first letter of a string.
function stringLib.ucFirst(s)
  return s:gsub("^%l", string.upper)
end

-- lcFirst :: string -> string
-- Lower-cases the first letter of a string.
function stringLib.lcFirst(s)
  return s:gsub("^%u", string.lower)
end

-- upperCaseWordsHelper :: string -> string -> string
-- Capitalizes x, and concatenates with xs.
local function upperCaseWordsHelper(x, xs)
  return x:upper() .. xs:lower()
end

-- ucWords :: string -> string
-- Capitalizes every word in a string.
function stringLib.ucWords(s)
  return s:gsub("(%a)([%w_']*)", upperCaseWordsHelper)
end

-- underscoreToSpace :: string -> string
-- Replaces every occurance of "_" by a single whitespace.
function stringLib.underscoreToSpace(s)
  return s:gsub("_", " ")
end

-- doubleUnderscoreToApostrophe :: string -> string
-- Replaces every occurance of "__", two underscores, by "'", an apostrophe.
function stringLib.doubleUnderscoreToApostrophe(s)
	return s:gsub("__", "'")
end

-------------------------------------------------------------------------------------
-- | Fixing Words in strings |-------------------------------------------------------
-------------------------------------------------------------------------------------

-- fixGender :: Gender -> string -> string
-- Replaces occurances of xir-, xim- and xhe-macros by their gender-specific counterparts.
function stringLib.fixGender(gender, s)
  return
		stringLib.xir(gender, 
		stringLib.xim(gender,
		stringLib.xhe(gender, s)))
end

-- inverseFixGender :: string -> Gender
-- Replaces occurances of gender-specific words to their xir-, xim- or xhe-macro counterparts.
function stringLib.inverseFixGender(s)
  return
		stringLib.inverseXir(
		stringLib.inverseXim(
		stringLib.inverseXhe(s)))
end

-- xir :: Gender -> string -> string
-- Replaces occurances of xir-macros by their gender-specific counterpart.
function stringLib.xir(gender, s)
  if     gender ==        "male" then return s:gsub("%%xir",        "his") 
  elseif gender ==      "female" then return s:gsub("%%xir",        "her")
  elseif gender == "androgynous" then return s:gsub("%%xir", "his or her")
  else                                return s:gsub("%%xhe",        "its")
	end
end

-- inverseXir :: string -> Gender
-- Replaces occurances of "his", "her" or "his or her" by their xir-macro counterpart.
function stringLib.inverseXir(s)
  local output = s
  output = output:gsub(       " his ", " %%xir ")
  output = output:gsub(       " her ", " %%xir ")
  output = output:gsub(" his or her ", " %%xir ")
  return output
end

-- xim :: Gender -> string -> string
-- Replaces occurance of xim-macros by their gender-specific counterpart.
function stringLib.xim(gender, s)
  if     gender ==        "male" then return s:gsub("%%xim",        "him") 
  elseif gender ==      "female" then return s:gsub("%%xim",        "her")
  elseif gender == "androgynous" then return s:gsub("%%xim", "him or her")
  else                                return s:gsub("%%xhe",         "it")
	end
end

-- inverseXim :: string -> Gender
-- Replaces occurances of "him", "her" or "him or her" by their xim-macro counterpart.
function stringLib.inverseXim(s)
  local output = s
  output = output:gsub(       " him ", " %%xim ")
  output = output:gsub(       " her ", " %%xim ")
  output = output:gsub(" him or her ", " %%xim ")
  return output
end

-- xhe :: Gender -> string -> string
-- Replaces occurance of xhe-macros by their gender-specific counterpart.
function stringLib.xhe(gender, s)
  if     gender ==        "male" then return s:gsub("%%xhe",        "he") 
  elseif gender ==      "female" then return s:gsub("%%xhe",       "she")
  elseif gender == "androgynous" then return s:gsub("%%xhe", "he or she")
  else                                return s:gsub("%%xhe",        "it")
	end
end

-- inverseXim :: string -> Gender
-- Replaces occurances of "he", "she" or "he or she" by their xhe-macro counterpart.
function stringLib.inverseXhe(s)
  local output = s
  output = output:gsub(        " he ", " %%xhe ")
  output = output:gsub(       " she ", " %%xhe ")
  output = output:gsub( " he or she ", " %%xhe ")
  return output
end

-- aOrAn :: string -> string
-- Prefixes a word with either "a" or "an" based on the word.
function stringLib.aOrAn(word)
  if string.find(string.sub(word, 1,1), "[AEIOUaeiou]") then return "an"
  else return "a" end
end

-- trySingular :: string --> string
-- Removes the last letter of a word to try and make it singular.
function stringLib.trySingular(word)
	return string.sub(word, 1, string.len(word) - 1)
end

-------------------------------------------------------------------------------------
-- | Util |--------------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- getIncremental :: int -> (() -> int)
-- Returns a function that increases it's return value by 1 every time it is called and starts at the parameter.
function util.getIncremental(start)
  local i = start
  return function()
    i = i + 1
    return i - 1
  end
end

-- log :: a -> a
-- Prints a value to the chat and returns the value.
function util.log(x)
  if NPCG_Options.debug == true then
    print(x)
  end
  return x
end

-- logTable :: { k -> a } -> { k -> a }
-- Prints all keys and values of a table to the chat and returns the table.
function util.logTable(t)
  for k,v in pairs(t) do
    print("(" .. k .. ", " .. v .. ")")
  end
  return t
end

-- logTableMapped :: { a -> b } -> { k -> a } -> { k -> b }
-- Prints all keys and values of a table after mapping a function over it to the chat and returns the (mapped) table.
function util.logTableMapped(f, t)
  return util.logTable(functua.table.map(f, t))
end

-- logTableOfPropertyLists :: { k -> { vk -> v } } -> { k -> { vk -> v } }
-- Prints all keys and their property keys and values to the chat and returns the input.
function util.logTableOfPropertyLists(t)
  for k,vs in pairs(t) do
    for vk, v in pairs(vs) do
      print(k .. " = (" .. vk .. ", " .. v .. ")")
    end
  end
  return t
end

-- logArray :: { int -> a } -> { int -> a }
-- Prints the array part of a table to the char and returns the table.
function util.logArray(t)
  for i,v in ipairs(t) do
    print("(" .. i .. ": " .. v .. ")")
  end
  return t
end

-- eitherOr :: a -> a -> a
-- Returns either the first or the second argument randomly.
function util.eitherOrRandom(x, y)
  return util.eitherOr(random(1,2) == 1, x, y)
end

-- eitherOrP :: bool -> a -> a -> a
-- Returns either the second argument if p is true, or the third argument otherwise.
function util.eitherOr(p, x, y)
  if p then
    return x
  else
    return y
  end
end

-- switch :: bool -> a -> ... -> a
-- Takes a number of booleans and cases, and returns the first case where the boolean before it is true. 
function util.switch(...)
  local xs = {...}
  
  local first = nil
  local setFirst = true
  
  for i,x in ipairs(xs) do
    if     setFirst then 
      first    = x      -- Found the predicate.
      setFirst = false  -- Swap case to read as value.
    elseif    first then 
      return x        -- Found predicate is true! Return value x.
    else
      setFirst = true -- Predicate is false, capture the next predicate.
    end
  end
end

-------------------------------------------------------------------------------------
-- | Random | -----------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- randomFrom :: { a } -> a
-- Returns a random element from a list.
-- Implicitly, if list is empty, nil will be returned.
function util.randomFrom(list)
	local c = functua.list.count(list)
	for _,v in pairs(list) do
		rand = random(1, c)
		if rand == 1 then
			return v 
		else
			c = c - 1
		end
	end
end

-- weightedRandomFrom :: { a } -> (a -> int) -> a
-- Returns a random element from a list using a weight function.
-- Implicitly, if the list is empty, nil will be returned.
function util.weightedRandomFrom(list, weightFunction)
  local sumOfWeights = functua.table.sumBy(list, weightFunction)
  if sumOfWeights == 0 then return nil end

  -- If we consider items to be duplicated by their weight in a fictive list we can take the item at the random index.
  local randomIndex = random(1, sumOfWeights)
  
  local foundKey = functua.table.findKey(
    function(x)
      if    x >= randomIndex  then  return true
      else                    randomIndex = randomIndex - x
                              return false end
    end,
    functua.table.map(weightFunction, list)
  )
    
  return list[foundKey]
end

-- uniqueRandomsFrom :: int -> { a } -> { a }
-- Returns a number between 1 and maxReturns of random elements from a list.
function util.uniqueRandomsFrom(maxReturns, list)
  if maxReturns <= 0 then return { } end
  
  local output = { }
	local c = functua.count(list)
	for _,v in pairs(list) do
		rand = random(1, c)
		if rand == 1 then
			table.insert(output, v) -- This line is changed to hopefully fix the bug >.>
      maxReturns = maxReturns - 1
      if maxReturns <= 0 then
        return output
      end
		else  -- Else clause jumbles up the c variable, this guarantees that there could be less returns than maxReturns.
			c = c - 1
		end
	end
  return output
end

-- uniqueRandomsFromCartesian :: { a } -> { b } -> a,b
-- Returns a random element of the first list, and one of the second list, ensuring they are not the same.
-- (x, y) | x <- xs and y <- ys and x ~= y 
function util.uniqueRandomsFromCartesian(xs, ys)
  local x = util.randomFrom(xs)
  local y = util.randomFrom(ys)
  while x == y and y ~= nil do -- This can loop infinitely.
    y = util.randomFrom(ys)
  end
  return x, y
end

-------------------------------------------------------------------------------------
-- | Clients |-----------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- getExpansionLevel :: string
-- Returns the current expansion level of the game.
function util.getExpansionLevel()
  local tocVersion = select(4, GetBuildInfo())
  local versionTable = {
    { minToc =     0, expansion =        "Warcraft..?" },
    { minToc = 10000, expansion =            "Classic" },
    { minToc = 80000, expansion = "Battle for Azeroth" },
    { minToc = 90000, expansion =        "Shadowlands" },
  }
  
  local version = functua.list.foldr(
    function(output, val)
      if   tocVersion > val.minToc
      then return    val
      else return output end 
    end,
    { minToc = 0, expansion = "Orcs vs Humans" },
    versionTable)
  return version.expansion
end

-- isClassic :: bool
-- Returns whether the current expansion level is "Classic".
function util.isClassic()
  return util.getExpansionLevel() == "Classic"
end

-- eitherOrByExpansionLevel :: a -> a -> a -> a
-- Returns the first, second or third input based on the expansion level.
function util.eitherOrByExpansionLevel(classicOutput, retailOutput, nextOutput)
  local expansionLevel = util.getExpansionLevel()
  
  if      expansionLevel == "Classic"             then  return classicOutput
  elseif  expansionLevel == "Battle for Azeroth"  then  return  retailOutput
  else                                                  return    nextOutput
  end
end

-- End of File.