-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local stringLib = npcg.functua.string
local util 			= npcg.functua.util

local GetCurrentCharacterTemplate = npcg.system.getCurrentCharacterTemplate

-------------------------------------------------------------------------------------
-- | Constants | --------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- These are Trait Definitions and will perform the basis for future stuff.
local NAME 				= { title = "Name", 						name = "Name", 						hasMultipleFields = false, requiredForPrinting =  true, description = { "The name of the NPC is based on the Race and Gender traits" } }
local RACE 				= { title = "Race", 						name = "Race", 						hasMultipleFields = false, requiredForPrinting =  true, description = { "The race of the NPC is picked from the playable and lorewise appropriate races. Location and faction decide the outcome of this trait, but these dependencies can be changed in the options" } }
local TRIBE 			= { title = "SubRace",					name = "Tribe",						hasMultipleFields = false, requiredForPrinting =  true, description = { "The tribe or subrace of the NPC and it's meaning depend on the race of the NPC. Location and faciton play a part in deciding on this trait, but this can be modified in the options." } }
local GENDER 			= { title = "Gender", 					name = "Gender", 					hasMultipleFields = false, requiredForPrinting =  true, description = { "The gender of the NPC is either male, female or androgynous (being hard to determine the gender on, or having both masculine and feminine characteristics" } }
local AGE 				= { title = "Age", 							name = "Age", 						hasMultipleFields = false, requiredForPrinting =  true, description = { "The age of the NPC described in approximations rather than years, as the meaning of these vary per race" } }
local ALIGNMENT 	= { title = "Alignment", 				name = "Alignment", 			hasMultipleFields = false, requiredForPrinting = false, description = { "The alignment of the NPC, based on \"DND\" alignments, describing the way the NPC is expected to act" } }
local FAITH 			= { title = "Faith", 						name = "Faith", 					hasMultipleFields = false, requiredForPrinting = false, description = { "The religion or faith that drives the npc." } }
local CLASS 			= { title = "Class", 						name = "Class", 					hasMultipleFields = false, requiredForPrinting =  true, description = { "The class of the NPC indicates the job for the NPC. This includes the playable classes (excluding Death Knight and Demon Hunter, which function more like races) and also includes other possible jobs" } }
local FEATURE 		= { title = "At First Glance",	name = "Feature",					hasMultipleFields = false, requiredForPrinting =  true, description = { "Something that immediately stands out for the NPC. It can be considered as a hook for observers to react on" } }
local MOTIVATIONS = { title = "Motivations", 			name = "Motivations", 		hasMultipleFields =  true, requiredForPrinting = false, description = { "Motivations that drive the NPC. (This is under redesign)" } }


local TRAITS = { Name=NAME, Race=RACE, Tribe=TRIBE, Gender=GENDER, Age=AGE, Alignment=ALIGNMENT, Faith=FAITH, Class=CLASS, Feature=FEATURE, Motivations=MOTIVATIONS }
local TRAITS_ORDERED = { RACE, TRIBE, GENDER, AGE, ALIGNMENT, FAITH, CLASS, FEATURE, MOTIVATIONS, NAME }

-- Height of the frame when not minimized.
local MAXIMIZED_HEIGHT = 422
-- Height of the frame when minimized.
local MINIMIZED_HEIGHT = 106

-------------------------------------------------------------------------------------
-- | Frame State | ------------------------------------------------------------------
-------------------------------------------------------------------------------------

local minimized = false

-------------------------------------------------------------------------------------
-- | Initialization Functions |------------------------------------------------------
-------------------------------------------------------------------------------------

local function InitializeTraitFrames(traitFrameContainer)
	-- Set Labels
	functua.table.forEach(
		function(trait)
				-- Set Title
				traitFrameContainer.Traits[trait.name]:SetTitle(trait.title)
				-- Set Tooltip
				traitFrameContainer.Traits[trait.name]:SetScript(								  "OnEnter", function(self) SetHelperInfo(trait.title, self, unpack(trait.description)) end)
			if not trait.hasMultipleFields then	
				-- RollButton Tooltip
				traitFrameContainer.Traits[trait.name].RollButton:SetScript(			"OnEnter", function(self) SetHelperText("Generate a new " .. trait.title, self) end)
				-- ClearButton Tooltip
				traitFrameContainer.Traits[trait.name].ClearButton:SetScript(		  "OnEnter", function(self) SetHelperText("Clear the "      .. trait.title, self) end)
				-- RollButton Click
				traitFrameContainer.Traits[trait.name].RollButton:SetScript(			"OnClick", function(self) npcg.system.rollTrait(  TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name) end)
				-- ClearButton Click
				traitFrameContainer.Traits[trait.name].ClearButton:SetScript(  	  "OnClick", function(self) npcg.system.clearTrait( TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name) end)
				-- OnUserInput
				traitFrameContainer.Traits[trait.name].EditBox:SetScript(	  "OnTextChanged", function(self) npcg.system.readTrait(  TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name) end)
			else
				-- Trait has multiple fields.
				functua.iterate( -- Set 3 fields.
					function(i)
						-- RollButton Tooltip
						traitFrameContainer.Traits[trait.name][ "RollButton"..i]:SetScript(				"OnEnter", function(self) SetHelperText("Generate a new " .. trait.title, self) end)
						-- ClearButton Tooltip
						traitFrameContainer.Traits[trait.name]["ClearButton"..i]:SetScript(				"OnEnter", function(self) SetHelperText("Clear the "      .. trait.title, self) end)
						-- RollButton Click
						traitFrameContainer.Traits[trait.name][ "RollButton"..i]:SetScript(				"OnClick", function(self)  npcg.system.rollTrait(  TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name, i) end)
						-- ClearButton Click
						traitFrameContainer.Traits[trait.name]["ClearButton"..i]:SetScript(  			"OnClick", function(self)  npcg.system.clearTrait( TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name, i) end)
						-- OnUserInput
						traitFrameContainer.Traits[trait.name][    "EditBox"..i]:SetScript(	"OnTextChanged", function(self)  npcg.system.readTrait(  TRAITS, traitFrameContainer, GetCurrentCharacterTemplate(), trait.name, i) end)
					end, 1, 3)
			end
		end,
		TRAITS) -- This should not use a constant, instead pass it to the function.
    
    
    -- Hard Code trait variable trait descriptions for now.
    -- Parse alignment descriptions.
    traitFrameContainer.Traits["Alignment"]:SetScript(								  "OnEnter", function(self) 
      local alignment = npcg.system.getCurrentCharacterTemplate():getAlignment()
      if alignment == "" then SetHelperInfo(
       "Alignment", self, "The alignment of the NPC, based on \"DND\" alignments, describing the way the NPC is expected to act")
      else                    SetHelperInfo(
       "Alignment: " .. stringLib.ucWords(alignment), self, npcg.data.alignment.getAlignmentDescription(alignment))
      end
    end)
    
    -- Parse age descriptions.
    traitFrameContainer.Traits["Age"]:SetScript(								        "OnEnter", function(self) 
      local age = npcg.system.getCurrentCharacterTemplate():getAge()
      if age == "" then SetHelperInfo(
       "Age", self, "The age of the NPC described in approximations rather than years, as the meaning of these vary per race")
      else                    SetHelperInfo(
       "Age: " .. stringLib.ucWords(age), self, npcg.data.age.getAgeDescription(age))
      end
    end)
end

local function InitializeTabs(tabFrame)
  -- Hi, I'm empty, can you check why..?
end

-------------------------------------------------------------------------------------
-- | Trait Generation |--------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Generates an Npc from the saved npcs.
local function generateNpcFromExisting()
  
  local characterTemplate = npcg.system.getCurrentCharacterTemplate()
  local filterF = functua.const(true) -- Use this filter to exclude certain results based on locations, times, context, ect.
  
  characterTemplate:clear()
  
  if NPCG_Options.factionSettings.context and NPCG_Options.factionSettings.matching then
		filterF = function(template) return npcg.data.race.isRaceWithinContext(template.race, GetZoneText(), GetSubZoneText()) end
	elseif NPCG_Options.factionSettings.matching then
		filterF = function(template) return npcg.data.race.isRaceOfFaction(template.race) end
	end
		
  local template = 
    functua.util.randomFrom(
      functua.list.filter(
        filterF,
        savedNpcTemplates))
  
  if template then
    characterTemplate:fromExisting(template)
    npcg.io.WriteToFrame(characterTemplate)
  else
    generateNpcUsingRandomGenerators()
  end
end

-- Generates an Npc using the random generators defined in modules/.
local function generateNpcUsingRandomGenerators()
  functua.list.forEach(
    function(trait) 
      if not trait.hasMultipleFields then
        npcg.system.rollTrait(TRAITS, self, GetCurrentCharacterTemplate(), trait.name) -- Pass in TRAITS
      else
        functua.iterate(
          function(i)
            npcg.system.rollTrait(TRAITS, self, GetCurrentCharacterTemplate(), trait.name, i) -- Pass in Traits
          end, 1, 3)
      end
    end,
    TRAITS_ORDERED) -- Pass this.
end

-- Top level generation, this decides what specific way of generating an Npc is used.
local function generateNpc()
  if NPCG_Options.friendSettings.friends and (random() <= NPCG_Options.friendSettings.chance) then    generateNpcFromExisting()
  else                                                                                                generateNpcUsingRandomGenerators() end
end

-------------------------------------------------------------------------------------
-- | Events |------------------------------------------------------------------------
-------------------------------------------------------------------------------------

function npcFrame_OnLoad(self)
	local tabContainer = self.Tabs

	-- Set up the TabContextFrame
	tabContainer.ContextFrame = CreateFrame("Frame", "TabContextFrame", tabContainer, "TabRightClickMenuTemplate")

	tabContainer.tabButtons[1]:HookScript("OnClick", function(self, button, isDown)
		
		if button == "RightButton" and     isDown and not self:GetParent()["ContextFrame"]:IsVisible() then
			self:GetParent()["ContextFrame"]:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
			self:GetParent()["ContextFrame"]["GetTabIndex"] = function() return 1 end
			self:GetParent()["ContextFrame"]["CloseTabButton"]:SetScript("OnClick", function()
					self:GetParent():RemoveTab(1)
					self:GetParent()["ContextFrame"]:Hide()
				end)
			self:GetParent()["ContextFrame"]:Show()
		else
			self:GetParent()["ContextFrame"]:Hide()
		end
		
	end)
	
	-- Subscribe to OnTabAdded to 
	tabContainer.OnTabAdded:AddListener(function(event, listener, tabIndex, ...)
		
		-- Left Click: Hide, Right Click: Show.
		tabContainer.tabButtons[tabIndex]:HookScript("OnClick", function(self, button, isDown)
			local contextFrame = self:GetParent().ContextFrame
			local tabContainer = self:GetParent()
			
			-- Open Context Menu
			if button == "RightButton" and     isDown and not contextFrame:IsVisible() then
				-- Set which tab is clicked.
				contextFrame.GetTabIndex = function() return tabIndex end -- This seems iffy...
				
				-- Remove Tab
				contextFrame.CloseTabButton:SetScript("OnClick", function()
					tabContainer:RemoveTab(tabIndex)
					contextFrame:Hide()
				end)
				
				-- Show ContextFrame
				contextFrame:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
				contextFrame:Show()
			else -- LeftButton
				contextFrame:Hide()
			end
		end)
		-- OnHover
		tabContainer.tabButtons[tabIndex]:SetScript("OnEnter", function(self)
			SetHelperInfo(
				-- Title: <Name>
				npcg.functua.string.ucWords(
					npcg.functua.util.eitherOr(
						-- has NPC
						not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(tabIndex).name),
						GetCharacterTemplateInTab(tabIndex).name,
						"Unnamed npc #" .. tabIndex)),
				-- Item
				self,
				-- Subtext: <Race>, <Class>
				npcg.functua.util.eitherOr(
					not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(tabIndex).class) and not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(tabIndex).race),
					npcg.functua.string.ucWords(GetCharacterTemplateInTab(tabIndex).race .. " " .. GetCharacterTemplateInTab(tabIndex).class),
					nil))	end)
	end) -- End Addlistener

	tabContainer.AddTabButton:HookScript("OnClick", function()
		tabContainer.ContextFrame:Hide()
	end)

	tabContainer.AddTabButton:HookScript("OnEnter", function(self)
		SetHelperText("Open a new tab", self)
	end)
	
	tabContainer.tabButtons[1]:SetScript("OnEnter", function(self) 
		SetHelperInfo(
			npcg.functua.string.ucWords(
				npcg.functua.util.eitherOr(
					not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(1).name),
					GetCharacterTemplateInTab(1).name, "Unnamed npc #" .. 1)),
			self,
			npcg.functua.util.eitherOr(
				not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(1).class) and not npcg.functua.string.nilOrWhitespace(GetCharacterTemplateInTab(1).race),
				npcg.functua.string.ucWords(GetCharacterTemplateInTab(1).race .. " " .. GetCharacterTemplateInTab(1).class),
				nil)
			)
	end)
	
	self.BottomBar.ClearButton:SetScript("OnClick", function()
		functua.table.forEach(
			function(trait) 
				if not trait.hasMultipleFields then
					npcg.system.clearTrait(TRAITS, self, GetCurrentCharacterTemplate(), trait.name)
				else
					functua.iterate(
						function(i)
							npcg.system.clearTrait(TRAITS, self, GetCurrentCharacterTemplate(), trait.name, 4 - i)
						end, 1, 3)
				end
			end,
			TRAITS)
	end)
	
	self.BottomBar.GenerateButton:SetScript("OnClick", generateNpc)

  self.BottomBar.GenerateButton:HookScript("OnClick", function()
    if NPCG_Options.interface.maximizeOnGenerate and minimized then
      self:SetSize(self:GetWidth(), MAXIMIZED_HEIGHT)
      self.Traits:Maximize()
      minimized = false
    end
  end)
	
	self.BottomBar.PrintButton:SetScript("OnClick", function()
		npcg.io.WriteToEmote(GetCurrentCharacterTemplate())
	end)

  self.MinimizeButton:HookScript("OnClick", function()
    if minimized then
      self:SetSize(self:GetWidth(), MAXIMIZED_HEIGHT)
      self.Traits:Maximize()
    else
      self:SetSize(self:GetWidth(), MINIMIZED_HEIGHT)
      self.Traits:Minimize()
    end
    
    minimized = not minimized
  end)
	
	InitializeTraitFrames(self)
	
	npcGenerator.OnLoad()
	self:RegisterForDrag("LeftButton")
end

-------------------------------------------------------------------------------------
-- | Tooltip Functions |-------------------------------------------------------------
-------------------------------------------------------------------------------------

function SetHelperText(text, item)
	if NPCG_Options.interface.enableTooltips == true then	
		if item ~= nil then
			GameTooltip:SetOwner(item, "ANCHOR_RIGHT")
			GameTooltip:SetText(text, 1, 1, 1)
			GameTooltip:FadeOut()
		end
	end
end

function SetHelperInfo(title, item, ...)
	if NPCG_Options.interface.enableTooltips == true then	
		if item ~= nil then
			GameTooltip:SetOwner(item, "ANCHOR_RIGHT")
			GameTooltip:SetText(title, 1, 1, 1)
			
			local arg = {...}
				for i,_ in pairs(arg) do 
					GameTooltip:AddLine(arg[i], 1.00,	0.82,	0.00, 1)
				end
				
			GameTooltip:Show()
			GameTooltip:FadeOut()
		end
	end
end

function SetHelperTextWarning(text)

end

function SetHelperTextError(text)

end

function SetHelperTextOption(text, item)
	if NPCG_Options.interface.enableTooltips == true then	
		if item ~= nil then
			GameTooltip:SetOwner(item, "ANCHOR_RIGHT")
			GameTooltip:SetText(text, 1.00,	0.82,	0.00, 1)
			GameTooltip:FadeOut()
		end
	end
end

-- End of File.