-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util
local replicate = functua.replicate

local generator = { }
local data 			= { }
local tests 		= { }

-- generators and data do not exist yet
npcg.generators.race	= generator
npcg.data.race				= data
npcg.tests.race				= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Data Table, this will be loaded in externally. Do not use this directly, use the functions instead.
local defaultTable = {
	[	"blood elf"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"civilized",
			"elven",
			"magic",
      "retail", }, },
	[ "death knight"  ] = {
		weight = 1,
		traitTypes = {
			"horde",
			"alliance",
			"hero",
      "retail", }, },
	[	"draenei"				] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"tribal",
			"magic",
			"civilized",
      "retail", }, },
	[	"dwarf"					] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
      "classic",
      "retail", }, },
	[	"gnome"					] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
			"magic",
      "classic",
      "retail", }, },
	[	"goblin"				] = {
		weight = 1,
		traitTypes = {
			"horde",
			"neutral",
			"civilized",
      "classic",
      "retail", }, },
	[	"half elf"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"alliance",
			"neutral",
			"civilized",
			"elven",
			"magic",
			"halfbreed",
      "classic",
      "retail", }, },
	[	"half orc"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"alliance",
			"neutral",
			"halfbreed",
      "classic",
      "retail", }, },		
	[	"high elf"			] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
			"elven",
			"magic",
      "classic",
      "retail", }, },
	[ "human"					] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
			"magic",
      "classic",
      "retail", }, },
	[ "illidari"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"alliance",
			"elven",
			"magic",
			"hero",
      "retail", }, },
	[ "night elf"			] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
			"elven",
      "classic",
      "retail", }, },
	[ "nightborne"		] = {
		weight = 1,
		traitTypes = {
			"horde",
			"civilized",
			"elven",
			"magic",
      "retail", }, },
	[ "orc"						] = {
		weight = 1,
		traitTypes = {
			"horde",
			"tribal",
      "classic",
      "retail", }, },
	[ "pandaren"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"alliance",
			"neutral",
      "retail", }, },
	[ "tauren"				] = {
		weight = 1,
		traitTypes = {
			"horde",
			"tribal",
      "classic",
      "retail", }, },
	[ "troll"					] = {
		weight = 1,
		traitTypes = {
			"horde",
			"tribal",
      "classic",
      "retail", }, },
	[ "forsaken"			] = {
		weight = 1,
		traitTypes = {
			"horde",
			"civilized",
      "classic",
      "retail", } },
	[ "void elf"			] = {
		weight = 1,
		traitTypes = {
				"alliance",
				"civilized",
				"elven",
				"magic",
				"hero",
        "retail", }, },
	[ "worgen"				] = {
		weight = 1,
		traitTypes = {
			"alliance",
			"civilized",
      "retail", }, },
  [ "vulpera"       ] = {
    weight = 1,
    traitTypes = {
      "horde",
      "tribal",
      "retail", }, },
	}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable

-- default :: { Race, RaceConstraints } -> { Race, RaceConstraints }
-- Returns the default race distrubution template altered by entries within the given table.
local function default(replacements)
	local replacements = replacements or { }
	
	return functua.table.mapKV(
		function(raceKey, raceValue)
			return replacements[raceKey] or functua.table.clone(raceValue) end,
		defaultTable)
end

-- defaultFromWeightTable :: { Race, Int } -> { Race, TribeConstraints }
-- Returns the default race distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(replacements)
	local output = default({ })
	
	for race, value in pairs(replacements) do
		if output[race] ~= nil then
			output[race].weight = value
		end
	end
	
	return output
end

-- defaultFromWeightTable :: { Race, Int } -> { Race, RaceConstraints }
-- Returns the default race distribution template altered by a weight table. The default weight is 0.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function fromWeightTable(replacements)
	local output = default({ })
	
	output = functua.table.map(
		function(v) local x = functua.table.clone(v); x.weight = 0; return x end, -- Must clone, v is a table.
		output)
	
	for race, value in pairs(replacements) do
		if output[race] ~= nil then
			output[race].weight = value
		end
	end
	
	return output
end

-- Adds the neutral raceType to all races in a table.
local function neutral(races)
	return functua.table.map(
		function(x)
			local output = functua.table.clone(x)
			output.traitTypes = functua.list.insert(output.traitTypes, "neutral")
			return output
			end,
		races)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- The default race distribution template, this is called when there are no specifics for zones and/or subzones.
local races = default({ })

---------------------------------------------------------------------------------------------------------------------------------------
-- | Zone-based Generation | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local zonesToRaces = {

-- Horde Capitals
["Orgrimmar"      		] = defaultFromWeightTable({
	["goblin"			] = 4,
	["orc"				] = 22,
	["pandaren" 	] = 2,
	["tauren"			] = 2,
	["troll"			] = 6,
	["forsaken"		] = 9,
}),
["Silvermoon City"		] = defaultFromWeightTable({
	["blood elf"	] = 40,
	["half elf"		] = 2,
	["forsaken" 	] = 3,
	["nightborne"	] = 6,
}),
["Thunder Bluff"			] = defaultFromWeightTable({
	["tauren"			] = 40,
	["troll"			] = 5,
}),

-- Alliance Capitals
["Stormwind City"			] = defaultFromWeightTable({
	["dwarf"			] = 5,
	["half elf"		] = 3,
	["gnome"			] = 3,
	["high elf"		] = 4,
	["human"			] = 30,
	["night elf"	] = 8,
	["worgen"			] = 3,
}),
["City of Ironforge"	] = defaultFromWeightTable({
	["dwarf"			] = 20,
	["gnome"			] = 10,
}),
["The Exodar"					] = defaultFromWeightTable({
	["draenei"		] = 40,
}),

-- Alliance Starting Zones

-- Azuremyst Isle
["Azuremyst Isle"			] = defaultFromWeightTable({
	["draenei"		] = 100,
}),
["Ammen Vale"					] = defaultFromWeightTable({
	["draenei"		] = 100,
}),

-- Dun Morogh
["Dun Morogh"					] = defaultFromWeightTable({
	["dwarf"  		] = 80,
	["gnome"			] = 20,
}),
["Coldridge Valley"		] = defaultFromWeightTable({
	["dwarf"  		] = 90,
	["gnome"			] = 10,
}),
["New Tinkertown"			] = defaultFromWeightTable({
	["dwarf"  		] = 10,
	["gnome"			] = 90,
}),

-- Elwynn Forest
["Elwynn Forest"			] = defaultFromWeightTable({
	["human"  		] = 100,
}),
["Northshire Valley"	] = defaultFromWeightTable({
	["human"  		] = 100,
}),

-- Horde Starting Zones

-- Durotar
["Durotar"						] = defaultFromWeightTable({
	["forsaken"		] = 10,
	["orc"				] = 30,
	["troll"			] = 10,
	["tauren"			] = 3,
}),
["Valley of Trials"		] = defaultFromWeightTable({
	["orc"				] = 100,
}),
["Echo Isles"					] = defaultFromWeightTable({
	["troll"			] = 100,
}),

-- Mulgore
["Mulgore"						] = defaultFromWeightTable({
	["tauren"			] = 100,
}),
["Camp Narache"				] = defaultFromWeightTable({
	["tauren"			] = 100,
}),

-- Eversong Woods
["Eversong Woods"			] = defaultFromWeightTable({
	["blood elf"	] = 100,
	["forsaken" 	] = 10,
	["half elf"		] = 2,
}),
["Falconwing Square"	] = nil, -- Tavern in Falconwing
["Ruins of Silvermoon"] = nil, -- Sometimes replaces Eversong Woods.
["Sunstrider Isle"		] = defaultFromWeightTable({
	["blood elf"	] = 100,
}),

-- Neutral
["Dalaran"        		] = neutral(defaultFromWeightTable({
	["blood elf"	] = 10,
	["gnome"			] = 10,
	["half elf"		] = 5,
	["high elf"		] = 15,
	["human"			] = 25,
	["nightborne"	] = 5,
})),

-- Warfronts
["Arathi Highlands"		] = neutral(defaultFromWeightTable({
	["blood elf"	] = 5,
	["forsaken"		] = 5,
	["human"			] = 4,
	["high elf"		] = 2,
	["dwarf"			] = 4,
	["gnome"			] = 2,
	["worgen"			] = 5,
})),
["Darkshore"					] = neutral(defaultFromWeightTable({
	["night elf"	] = 20,
})),

-- Battle for Azeroth

-- Zandalar
["Dazar'alor"					] = defaultFromWeightTable({ -- Dazar'alor is only a select number of subzones.
	["troll"			] = 40, }),
["Nazmir"							] = defaultFromWeightTable({
	["troll"			] = 40, }),
["Vol'dun"						] = defaultFromWeightTable({
	["troll"			] = 40, }),
["Zuldazar"						] = defaultFromWeightTable({
	["troll"			] = 40, }),

-- Kul'tiras
["Boralus" 						] = defaultFromWeightTable({
	["human"			] = 40, }),
["Drustvar"						] = defaultFromWeightTable({
	["human"			] = 40, }),
["Stormsong Valley"		] = defaultFromWeightTable({
	["human"			] = 40, }),
["Tiragarde Sound"		] = defaultFromWeightTable({
	["human"			] = 40, }),

}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Subzone-based Generation | -------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Make this local
local subzonesToRaces = {
	["The Cape of Stranglethorn"		] = {
		["Booty Bay"									] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})),
		["The Old Port Authority" 		] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})), },
	["Booty Bay"										] = {
		["The Salty Sailor Tavern"		] = neutral(defaultFromWeightTable({
			["goblin"			] = 20,
		})), },
	["Eversong Woods"								] = {
		-- Ruins of Silvermoon.
		["Commons Hall"								] = nil, -- Big house in the back of the Ruins.
		["Dawning Lane"								] = nil,
		["Feth's Way"									] = nil, -- Right Road in Ruins.
		["Falconwing Square"					] = nil,
		["Ruins of Silvermoon"				] = nil,
		["Skulking Row"								] = nil, -- Left Road in Ruins.
	},
	["Mardum, the Shattered Abyss"	]	= {
		["The Fel Hammer"							] = neutral(fromWeightTable({
			["illidari"   ] = 1,
		})), },
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Maybe abstract this part??

local function checkAllLocations(zone, subZone)
  return 
    -- Test whether the SubZone has values.
    ((subzonesToRaces[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function checkZonesOnly(zone)
  return
    -- Test whether the Zone has values.
    zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function dontCheckLocations()
  return
    -- Fallback to default.
    races -- zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- Race -> Zone -> SubZone -> bool

local function isClassicRace(race)
  return functua.list.contains("classic", race.traitTypes)
end

local function isRetailRace(race)
  return functua.list.contains("retail", race.traitTypes)
end

local function isNextExpansionRace(race)
  return functua.list.contains("future", race.traitTypes)
end

local function isRaceOfFaction(race)
	return
    functua.table.any(
      function(x) return x == string.lower(UnitFactionGroup("player")) or x == "neutral" end, 
      race.traitTypes)
end

-- This looks like it will backfire massively...
local function isRaceWithinContext(race, zone, subZone)
  return isRaceOfFaction(race)
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- { race } -> race
-- This might change in the future.

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.table.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Queries | ------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- allRacesExcept :: Race -> ... -> { Race }
function data.allRacesExcept(...)
  return functua.table.without(functua.table.keys(races), {...})
end

-- racesByType :: RaceType -> { Race }
function data.racesByType(raceType)
  return functua.table.filterKeysByValue(
    function(x) return
      functua.table.contains(raceType, x.traitTypes)
    end,
    util.eitherOr(raceType == "any", { }, races)
  )
end

-- racesByFaction :: Faction -> { Race }
function data.racesByFaction(faction)
  return data.racesByType(string.lower(faction))
end

-- REDO THIS
-- isRaceOfFaction :: Race --> bool
function data.isRaceOfFaction(race)
	local raceWithConstraints = (	
    -- Fallback to default.
    races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { })[race]
	
	return isRaceOfFaction(raceWithConstraints or {traitTypes={"neutral"}})
end

-- REDO THIS
-- isRaceWithinContext :: Race --> Zone --> Subzone --> bool
function data.isRaceWithinContext(race, zone, subZone)
	local raceWithConstraints = (
		-- Test whether the SubZone has values.
    ((subzonesToRaces[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToRaces[zone]
    -- Fallback to default.
    or races --zonesToRaces["Default"]
    -- Fallback in case some extreme breaking happened.
    or { })[race]	
	return isRaceWithinContext(raceWithConstraints or {traitTypes={"neutral"}}, zone, subZone)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local function createRaceGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone) 
    return selectorF(
			functua.table.filterV(
				function(race) return filterF(race, zone, subZone) end, 
				generatorF(zone, subZone))) or "Error: No options found!"
  end
end

-- Returns the generatorF argument for createRaceGenerator based on the current options set in NPCG_Options.locationSettings
local function getGeneratorFunction()
  if     not NPCG_Options.locationSettings.zones then
    return dontCheckLocations
  elseif not NPCG_Options.locationSettings.subZones then
    return checkZonesOnly
  else
    return checkAllLocations
  end
end

-- Returns the filterF argument for createRaceGenerator based on the current options set in NPCG_Options.factionSettings and the Game Level.
local function getFilterFunction()
  local output = functua.const(true)
  
  -- Check faction settings.
  if      not NPCG_Options.factionSettings.matching then
    output = functua.const(true)
  elseif  not NPCG_Options.factionSettings.context then
    output = isRaceOfFaction
  else
    output = isRaceWithinContext
  end
  
  -- Check Expansion Level.
  return functua.both(output, util.eitherOrByExpansionLevel(isClassicRace, isRetailRace, isNextExpansionRace))
end

-- This function attempts to set the race generator based on settings.
function generator.initialize()
  generateRandomRace = function() return "Error: generateRandomRace is not set" end
  
  -- Arguments for createTraitGenerator
  local generatorF  = getGeneratorFunction()
  local filterF     = getFilterFunction()
  local selectorF   = selectUsingWeightedRandom
  
  generateRandomRace = createRaceGenerator(generatorF, filterF, selectorF)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Exposing Functions and Data | -----------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

data.defaultTable     = races
data.zoneDataTable    = zonesToRaces
data.subZoneDataTable = subzonesToRaces

---------------------------------------------------------------------------------------------------------------------------------------
--| Testing Functions | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

function tests.testZoneToRaceRatio()
  local test = { }
  local zone   =  util.eitherOr(
                    NPCG_Options.locationSettings.mock  and 
                    NPCG_Options.locationSettings.zones, 
                    NPCG_Options.locationSettings.mockZone,
                    GetZoneText())
  local subzone = util.eitherOr(
                    NPCG_Options.locationSettings.mockSub and
                    NPCG_Options.locationSettings.zones   and
                    NPCG_Options.locationSettings.subZones,
                    NPCG_Options.locationSettings.mockSubZone or "<No SubZone>",
                    GetSubZoneText() or "<No SubZone>")
  print("Testing Zone to Race Ratio for: " .. zone .. ", " .. subzone)
  functua.list.forEach(
    function(x) 
      local result = generateRandomRace(nil, zone, subzone)
      test[result] = (test[result] or 0) + 1 end,
    functua.replicate(10000, 1))
  util.logTableMapped(function(x) return x .. ", (" .. ((x / 10000) * 100) .. "%)" end, test)
end


-- End of File.